/***********************************************************************
 * File: buffer_checker.c
 * (Extracted from
 * "$Id: buffer_checker.c,v 1.13 2006/04/08 14:43:08 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source code implements a buffer checker using regex(3).
 *
 * Copyright (C) 2004 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include <stdio.h>
#include <buffer_checker.h>


/* declare a list of regular-expression and processed-expression pairs.
 */
static struct regexe_preg_list_t {
	struct regex_preg_pair_t {
		char	*regex;	/* pointing to a regular expression */
		regex_t	*preg;	/* pointing to a processed expression
				 * that is allocated memory.
				 */
	}			*pairs;
	ssize_t			 pairs_cnt;
}			regex_preg_list = { NULL, 0 };


/* Utility for this library:
 * find the pair of regular expression and processed expression that
 * is matched a regular expression of the entry of `Buffer_checker_t'.
 */
static ssize_t find_pair(Buffer_checker_t *define_ent)
{
	ssize_t	i;

	/* test the argument */
	if (define_ent == NULL) {
		return	BCR_Illegal_args;
	}

	for (i = 0; i < regex_preg_list.pairs_cnt; ++i) {
		if (strcmp(regex_preg_list.pairs[i].regex
				, define_ent->regex
				) == 0) {
			return	i; /* Found the pair */
		}
	}

	return	(ssize_t)-1; /* Not found the pair */
}


/* Utility of this library:
 * compile the new regular expression (and allocate several memories)
 */
static Buffer_checker_result_t compile_regex(
					Buffer_checker_t *define_ent
					)
{
	ssize_t	idx;

	/* test the argument */
	if (define_ent == NULL) {
		return	BCR_Illegal_args; /* Illegal argument! */
	}

	/* test was this entry compiled, yet? */
	if (define_ent->preg != NULL) {
		/* This entry was already compiled. */
		return	BCR_Continue;
	}

	if ((regex_preg_list.pairs_cnt == 0)
	|| (idx = find_pair(define_ent)) < 0) {
		/* (re-)allocate pairs of the regular-expression and
		 * processed-expression pairs
		 */
		idx	= ++regex_preg_list.pairs_cnt;
		if (((regex_preg_list.pairs =
			(struct regex_preg_pair_t *)realloc(
				  regex_preg_list.pairs
				, sizeof(*regex_preg_list.pairs) * idx
				)) == NULL)
		|| (regex_preg_list.pairs[idx - 1].preg =
			(regex_t *)malloc(sizeof(regex_t))) == NULL) {
			return	BCR_Cannot_allocate_preg;
		}

		/* set the regular expression of this entry */
		regex_preg_list.pairs[idx - 1].regex
						= define_ent->regex;

		/* compile the regular expression, and set it */
		if (regcomp(regex_preg_list.pairs[idx - 1].preg
				, define_ent->regex
				, REG_EXTENDED | REG_NOSUB
				) != 0) {
			return	BCR_Failed_to_compile_regex;
		}
		define_ent->preg = regex_preg_list.pairs[idx - 1].preg;
	}
	else {
		define_ent->preg = regex_preg_list.pairs[idx].preg;
	}

	return	BCR_Continue;	/* Normal End */
}


/* API: For debugging, dump the `regex_preg_list' structure.
 *      `debug' argument is not zero, this API prints it to stderr.
 */
void dump_buffer_checker_regex_preg_list(int debug)
{
	ssize_t	i;

	/* rest the argument */
	if (debug == 0)	 {
		return;	/* Zero means non-debugging mode */
	}

	fprintf(stderr, ">>> %s: regex_preg_list contains...\n"
	,		__FILE__); fflush(stderr);
	fprintf(stderr, "\tpairs_cnt = %ld.\n"
		,		(long)(regex_preg_list.pairs_cnt)
	); fflush(stderr);
	for (i = 0; i < regex_preg_list.pairs_cnt; ++i) {
		fprintf(stderr
		, "\tpairs[%ld].regex = \"%s\" "
				"(%p)\n"
		  "\t          .preg  = %p\n"
		,	(long)i, regex_preg_list.pairs[i].regex
		,		regex_preg_list.pairs[i].regex
		,		(void *)(regex_preg_list.pairs[i].preg)
		); fflush(stderr);
	}

	return;
}

/* API: Initialize the `Buffer_checker_t' structure array.
 *	This API must be invoked once, at the first in the main thread.
 */
Buffer_checker_result_t init_buffer_checker(Buffer_checker_t *defines
					, size_t	defines_cnt
					, size_t	buffer_sz
					)
{
	Buffer_checker_result_t	result;
	size_t	i, flds_sz = 0;

	/* test the arguments */
	if (((defines == NULL) || defines_cnt <= 0) || buffer_sz <= 0) {
		return	BCR_Illegal_args;
	}

	/* initialize the each regular expressions... */
	for (i = 0; i < defines_cnt; ++i) {
		/* test the specifications of Buffer_checker_t array */
		if ((defines[i].fld_offset < 0)
		|| (defines[i].regex == NULL
			&& defines[i].cond_cb == NULL)
		) {
			return	BCR_Illegal_defines;
		}

		/* the regular expression isn't compiled? */
		if ((defines[i].regex != NULL)
		&& defines[i].preg == NULL) {
			if ((result = compile_regex(defines + i))
						!= BCR_Continue) {
				return	result;
			}
		}

		/* sum up the size of fields */
		flds_sz	+= defines[i].fldsz;
	}

	/* test the total size of fields */
	return	BCR_Continue; /* Normal End */
}


/* API: Check the all fields in a buffer, using initialized array. */
Buffer_checker_result_t exec_buffer_checker(void *buffer
					, size_t	buffer_sz
					, Buffer_checker_t *defines
					, size_t	defines_cnt
					, void		*user_cntxt
					, size_t	*executed_cnt
					)
{
	int	result	= BCR_Continue;
	size_t	i;
	char	*buffer_last_addr = NULL;
	ssize_t	clause_cbs_idx = -1;
	Buffer_checker_regex_result_t	regex_result;

	/* test the arguments */
	if ((((buffer == NULL
	|| buffer_sz <= 0)
	|| defines == NULL)
	|| defines_cnt <= 0)
	|| executed_cnt == NULL) {
		return	BCR_Illegal_args;
	}
	for (i = 0; i < defines_cnt; ++i) {
		if ((defines[i].preg == NULL)
		&& defines[i].cond_cb == NULL) {
			return	BCR_Broken_Configuration;
		}
	}

	/* check the each fields... */
	buffer_last_addr = ((char *)buffer) + buffer_sz;
	for (*executed_cnt = 0, i = 0
	; result == (int)BCR_Continue && i < defines_cnt
	; ++(*executed_cnt)) {
		char	*fldp = NULL;
		size_t	fldsz	= defines[i].fldsz;

		/* get the address of the field */
		if (defines[i].user_get_fld_addr != NULL) {
			/* invoke the user specified function */
			if ((fldp = (*(defines[i].user_get_fld_addr))(
						  buffer
						, defines[i].fld_offset
						, &fldsz
						, user_cntxt
						)) == NULL) {
				return	BCR_Cannot_get_field_addr;
			}
		}
		else {	/* get address from the buffer */
			fldp = ((char *)buffer) + defines[i].fld_offset;
		}

		/* test size of the field is illegal? */
		if (fldsz == 0) {
			return	BCR_Field_size_is_illegal;
		}

		/* test the field runs over the buffer? */
		if ((fldp < ((char *)buffer))
		|| (fldp + fldsz) > buffer_last_addr) {
			return	BCR_Field_runs_over_buffer;
		}

		/* test the length of the field-string */
		if ((defines[i].regex != NULL)
		&& strlen(fldp) >= fldsz) {
			return	BCR_Field_string_is_too_long;
		}

		/* execute the regex-matching function */
		regex_result	= ((defines[i].preg == NULL)?
					BCRR_Not_Specified_Regex:
					(regexec(defines[i].preg
						, fldp
						, (size_t)0
						, NULL
						, 0
						) == 0)?
							BCRR_Matched:
							BCRR_Unmatched
					);

		/* invoke the `cond_cb' callback */
		if (defines[i].cond_cb == NULL) {
			clause_cbs_idx	= regex_result;
		}
		else if (((clause_cbs_idx = (*(defines[i].cond_cb))(
							  i
							, buffer
							, buffer_sz
							, fldp
							, regex_result
							, user_cntxt
							)) < 0)
		|| clause_cbs_idx >= (ssize_t)(
			(sizeof(defines[0].clause_cbs)
				/ sizeof(defines[0].clause_cbs[0])))
		) {	/* `cond_cb' callback returns out of range
			 * (`clause_cbs[]' array)
			 */
			return	BCR_Out_of_range_cond_cb_returns;
		}

		/* invoke `clause_cbs[clause_cbs_idx]' callback */
		if (defines[i].clause_cbs[clause_cbs_idx] != NULL) {
			/* invoke the user's matched callback */
			result = (*(defines[i]
					.clause_cbs[clause_cbs_idx]))(
						  i
						, buffer
						, buffer_sz
						, fldp
						, user_cntxt
						);
		}
	
		/* test the result from the callbacks:
		 *   if it is equal or greater than 0, then
		 * let `i' `result' and let `result' `BCR_Continue'.
		 * Otherwise, increase `i'.
		 */
		if (result >= 0) {
			i	= result;
			result	= BCR_Continue;
		}
		else {
			++i;
		}
	}

	return	(Buffer_checker_result_t)result;
			/* As result from user's callbacks.
			 * BCR_Continue (all fields are sane)
			 * or other codes
			 */
}


/* API: Clean up the `Buffer_checker_t' structure array.
 *	This API must be invoked once, at last in the main thread.
 */
Buffer_checker_result_t	finish_buffer_checker(Buffer_checker_t *defines
						, ssize_t defines_cnt
						)
{
	ssize_t	i;

	/* test the arguments */
	if ((defines == NULL) || defines_cnt <= 0) {
		return	BCR_Illegal_args;
	}

	/* finalize the each regular expressions... */
	for (i = 0; i < defines_cnt; ++i) {
		/* the expression has been allocated? */
		if (defines[i].preg != NULL) {
			defines[i].preg	= NULL;
		}
	}

	/* release the memories around of processed-expressions */
	for (i = 0; i < regex_preg_list.pairs_cnt; ++i) {
		/* release the allocated compiled expression */
		regfree(regex_preg_list.pairs[i].preg);

		/* release the allocated regex_t memory */
		free(regex_preg_list.pairs[i].preg);
	}
	regex_preg_list.pairs_cnt	= 0;

	return	BCR_Continue;	/* Normal End */
}
